(function (global, factory) {
	typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
	typeof define === 'function' && define.amd ? define(['exports'], factory) :
	(factory((global.Viewprt = global.Viewprt || {})));
}(this, (function (exports) { 'use strict';

/**
 * @class `Viewport`
 * A a scrollable container containing multiple observers
 * that are checked each time the viewport is manipulated (scrolled, resized, mutated)
 */
function Viewport(container) {
  var _this = this;

  this.container = container;
  this.observers = [];
  this.lastY = 0;
  var element = this.element = container === document.body ? window : container;

  var scheduled = false;
  var throttle = window.requestAnimationFrame || function (callback) {
    return setTimeout(callback, 1000 / 60);
  };
  var handler = this.handler = function () {
    if (!scheduled) {
      scheduled = true;
      throttle(function () {
        var state = _this.getState();
        _this.checkObservers(state);
        _this.lastY = state.y;
        scheduled = false;
      });
    }
  };

  element.addEventListener('scroll', handler);
  element.addEventListener('resize', handler);

  if (window.MutationObserver) {
    addEventListener('DOMContentLoaded', function () {
      var mutationObserver = _this.mutationObserver = new MutationObserver(handler);
      mutationObserver.observe(container, { attributes: true, childList: true, subtree: true });
    });
  }
}

Viewport.prototype = {
  addObserver: function addObserver(observer) {
    var observers = this.observers;

    observers.indexOf(observer) === -1 && observers.push(observer);
  },
  removeObserver: function removeObserver(observer) {
    var observers = this.observers;

    var index = observers.indexOf(observer);
    index > -1 && observers.splice(index, 1);
  },
  checkObservers: function checkObservers(state) {
    var observers = this.observers;

    for (var i = observers.length; i--;) {
      observers[i].check(state);
    }
  },
  getState: function getState() {
    var element = this.element,
        lastY = this.lastY;

    var width = void 0,
        height = void 0,
        y = void 0;
    if (element === window) {
      width = element.innerWidth;
      height = element.innerHeight;
      y = element.pageYOffset;
    } else {
      width = element.offsetWidth;
      height = element.offsetHeight;
      y = element.scrollTop;
    }
    var yDirection = lastY < y ? 'down' : 'up';
    return { width: width, height: height, y: y, yDirection: yDirection };
  },
  destroy: function destroy() {
    var element = this.element,
        handler = this.handler,
        mutationObserver = this.mutationObserver;

    element.removeEventListener('scroll', handler);
    element.removeEventListener('resize', handler);
    mutationObserver && mutationObserver.disconnect();
  }
};

/**
 * When adding a new observer, the manager checks if the viewport
 * already exists so they share common event listeners.
 */
var viewports = [];

function addViewportObserver(observer) {
  var container = observer.container;
  var index = getViewportIndexForContainer(container);
  var viewport = void 0;

  if (index > -1) {
    viewport = viewports[index];
  } else {
    viewport = new Viewport(container);
    viewports.push(viewport);
  }

  viewport.addObserver(observer);
  return viewport;
}

function removeViewportObserver(observer) {
  var index = getViewportIndexForContainer(observer.container);
  var viewport = viewports[index];

  if (viewport) {
    viewport.removeObserver(observer);
    if (!viewport.observers.length) {
      viewport.destroy();
      viewports.splice(index, 1);
    }
  }
}

function getViewportIndexForContainer(container) {
  for (var i = viewports.length; i--;) {
    if (viewports[i].container === container) {
      return i;
    }
  }
}

/**
 * @interface Observer
 * Each type of observer implements these options/methods
 */
function Observer(opts) {
  this.offset = ~~opts.offset || 0;
  this.container = opts.container || document.body;
  this.once = Boolean(opts.once);
  return this.activate();
}

Observer.prototype = {
  activate: function activate() {
    return addViewportObserver(this);
  },
  destroy: function destroy() {
    removeViewportObserver(this);
  }
};

function ObserverInterface(Subclass) {
  Subclass.prototype = Object.create(Observer.prototype);
  Subclass.prototype.constructor = Subclass;
  return Subclass;
}

var PositionObserver = ObserverInterface(function PositionObserver() {
  var opts = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

  if (!(this instanceof PositionObserver)) {
    return new (Function.prototype.bind.apply(PositionObserver, [null].concat(Array.prototype.slice.call(arguments))))();
  }

  this.onTop = opts.onTop;
  this.onBottom = opts.onBottom;
  this.onMaximized = opts.onMaximized;
  this._wasTop = true;
  this._wasBottom = false;

  var viewport = Observer.call(this, opts);
  this.check(viewport.getState());
});

PositionObserver.prototype.check = function (viewportState) {
  var onBottom = this.onBottom,
      onTop = this.onTop,
      onMaximized = this.onMaximized,
      _wasTop = this._wasTop,
      _wasBottom = this._wasBottom,
      container = this.container,
      offset = this.offset,
      once = this.once;
  var scrollHeight = container.scrollHeight;
  var height = viewportState.height,
      y = viewportState.y;

  var atTop = y - offset <= 0;
  var atBottom = scrollHeight > height && height + y + offset >= scrollHeight;
  var untriggered = false;

  if (onBottom && !_wasBottom && atBottom) {
    onBottom.call(this, container, viewportState);
  } else if (onTop && !_wasTop && atTop) {
    onTop.call(this, container, viewportState);
  } else if (onMaximized && scrollHeight === height) {
    onMaximized.call(this, container, viewportState);
  } else {
    untriggered = true;
  }

  if (once && !untriggered) {
    this.destroy();
  }

  this._wasTop = atTop;
  this._wasBottom = atBottom;
};

function isElementInViewport(element, offset, viewportState) {
  var rect = element.getBoundingClientRect();
  return !!(rect.width && rect.height) && rect.top < viewportState.height + offset && rect.bottom > 0 - offset && rect.left < viewportState.width + offset && rect.right > 0 - offset;
}

function isElementInDOM(element) {
  return element && element.parentNode;
}

var ElementObserver = ObserverInterface(function ElementObserver(element) {
  var opts = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

  if (!(this instanceof ElementObserver)) {
    return new (Function.prototype.bind.apply(ElementObserver, [null].concat(Array.prototype.slice.call(arguments))))();
  }

  this.element = element;
  this.onEnter = opts.onEnter;
  this.onExit = opts.onExit;
  this._didEnter = false;
  var viewport = Observer.call(this, opts);

  if (isElementInDOM(element)) {
    this.check(viewport.getState());
  }
});

ElementObserver.prototype.check = function (viewportState) {
  var onEnter = this.onEnter,
      onExit = this.onExit,
      element = this.element,
      offset = this.offset,
      once = this.once,
      _didEnter = this._didEnter;


  if (!isElementInDOM(element)) {
    this.destroy();
  } else if (onEnter && !_didEnter && isElementInViewport(element, offset, viewportState)) {
    this._didEnter = true;
    onEnter.call(this, element, viewportState);
    once && this.destroy();
  } else if (onExit && _didEnter && !isElementInViewport(element, offset, viewportState)) {
    this._didEnter = false;
    onExit.call(this, element, viewportState);
    once && this.destroy();
  }
};

exports.PositionObserver = PositionObserver;
exports.ElementObserver = ElementObserver;

Object.defineProperty(exports, '__esModule', { value: true });

})));
